@@ -4,46 +4,43 @@
 //! * [Limine Boot Protocol Specification](https://github.com/limine-bootloader/limine/blob/trunk/PROTOCOL.md)
 
 #![no_std]
+#![feature(const_nonnull_new)]
 
 use core::cell::UnsafeCell;
 use core::fmt::Debug;
+use core::ptr::NonNull;
 
 #[derive(Debug)]
 #[repr(transparent)]
 pub struct LimineEntryPoint(*const ());
 
 #[repr(transparent)]
-pub struct LiminePtr<T: Debug>(*const T);
+pub struct LiminePtr<T: Debug>(Option<NonNull<T>>);
 
 impl<T: Debug> LiminePtr<T> {
-    const DEFAULT: LiminePtr<T> = Self(core::ptr::null_mut() as *const T);
+    const DEFAULT: LiminePtr<T> = Self(None);
 
-    /// Returns the raw pointer.
-    ///
-    /// # Safety
-    /// The returned pointer may-be null.
     #[inline]
-    unsafe fn raw_get(&self) -> *const T {
-        self.0
+    pub fn as_mut_ptr(&self) -> Option<*mut T> {
+        Some(self.0?.as_ptr())
     }
 
-    /// Retrieve the value of the pointer. Returns an optional value since the pointer
-    /// may be null.
-    pub fn get(&self) -> Option<&'static T> {
-        let raw_ptr = unsafe { self.raw_get() };
+    #[inline]
+    pub fn as_ptr(&self) -> Option<*const T> {
+        self.as_mut_ptr().map(|e| e as _)
+    }
 
-        if raw_ptr.is_null() {
-            None
-        } else {
-            unsafe { Some(&*raw_ptr) }
-        }
+    #[inline]
+    pub fn get(&self) -> Option<&'static T> {
+        Some(unsafe { self.0?.as_ref() })
     }
 }
 
 impl LiminePtr<char> {
     /// Converts the limine string pointer into a rust string.
-    pub fn to_string(&self) -> &'static str {
-        let mut ptr = unsafe { self.raw_get() } as *const u8;
+    pub fn to_string(&self) -> Option<&'static str> {
+        let mut ptr = self.as_ptr()? as *const u8;
+        let ptr_copy = ptr;
 
         // 1. Calculate the length of the string.
         let mut str_len = 0;
@@ -60,26 +57,26 @@ impl LiminePtr<char> {
         //
         // SAFETY: We know that the string is null terminated and that the length
         // is calculated correctly.
-        let slice = unsafe { core::slice::from_raw_parts(self.raw_get() as *const u8, str_len) };
+        let slice = unsafe { core::slice::from_raw_parts(ptr_copy, str_len) };
 
         // 3. Convert the slice to a rust string.
         //
         // SAFETY: Limine strings are ensured to have valid UTF-8.
-        unsafe { core::str::from_utf8_unchecked(slice) }
+        unsafe { Some(core::str::from_utf8_unchecked(slice)) }
     }
 }
 
 impl LiminePtr<LimineEntryPoint> {
     #[inline]
     pub const fn new(entry_point: fn() -> !) -> Self {
-        Self(entry_point as *const _)
+        Self(NonNull::new(entry_point as *mut _))
     }
 }
 
 impl<T: 'static + Debug> Debug for LiminePtr<T> {
     fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
         f.debug_tuple("LiminePtr")
-            .field(&format_args!("{:#x?}", unsafe { self.raw_get() }))
+            .field(&format_args!("{:#x?}", self.get()))
             .finish()
     }
 }
@@ -133,9 +130,9 @@ macro_rules! make_struct {
 
             // generate a getter method for each field:
             $($(#[$field_meta])* pub const fn $field_name(mut self, value: $field_ty) -> Self {
-				self.$field_name = value;
-				self
-			})*
+                self.$field_name = value;
+                self
+            })*
         }
 
         // maker trait implementations for limine request struct:
@@ -255,14 +252,17 @@ pub struct LimineFramebufferResponse {
     /// How many framebuffers are present.
     pub framebuffer_count: u64,
     /// Pointer to an array of `framebuffer_count` pointers to struct [`LimineFramebuffer`] structures.
-    pub framebuffers: LiminePtr<*const LimineFramebuffer>,
+    pub framebuffers: LiminePtr<LiminePtr<LimineFramebuffer>>,
 }
 
 impl LimineFramebufferResponse {
     pub fn framebuffers(&self) -> Option<&'static [LimineFramebuffer]> {
         self.framebuffers.get().map(|entry| unsafe {
-            core::slice::from_raw_parts(*entry, self.framebuffer_count as usize)
-        })
+            Some(core::slice::from_raw_parts(
+                entry.as_ptr()?,
+                self.framebuffer_count as usize,
+            ))
+        })?
     }
 }
 
@@ -282,16 +282,8 @@ pub struct LimineTerminal {
     pub framebuffer: LiminePtr<LimineFramebuffer>,
 }
 
-#[repr(transparent)]
-struct LimineTerminalWrite(extern "C" fn(*const LimineTerminal, *const u8, u64));
-
-impl Debug for LimineTerminalWrite {
-    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
-        f.debug_tuple("LimineTerminalWrite")
-            .field(&format_args!("{:#x}", self.0 as usize))
-            .finish()
-    }
-}
+type LimineTerminalWrite =
+    Option<unsafe extern "C" fn(terminal: *const LimineTerminal, string: *const u8, length: u64)>;
 
 #[repr(C)]
 #[derive(Debug)]
@@ -301,27 +293,30 @@ pub struct LimineTerminalResponse {
     /// How many terminals are present.
     pub terminal_count: u64,
     /// Pointer to an array of `terminal_count` pointers to struct `limine_terminal` structures.
-    pub terminals: LiminePtr<*const LimineTerminal>,
+    pub terminals: LiminePtr<LiminePtr<LimineTerminal>>,
     /// Physical pointer to the terminal `write()` function. The function is not thread-safe, nor
     /// reentrant, per-terminal. This means multiple terminals may be called simultaneously, and
     /// multiple callbacks may be handled simultaneously. The terminal parameter points to the
     /// [`LimineTerminal`] structure to use to output the string; the string parameter points to
     /// a string to print; the length paremeter contains the length, in bytes, of the string to print.
-    write: LiminePtr<Option<LimineTerminalWrite>>,
+    write_fn: LimineTerminalWrite,
 }
 
 impl LimineTerminalResponse {
     pub fn terminals(&self) -> Option<&'static [LimineTerminal]> {
         self.terminals.get().map(|entry| unsafe {
-            core::slice::from_raw_parts(*entry, self.terminal_count as usize)
-        })
+            Some(core::slice::from_raw_parts(
+                entry.as_ptr()?,
+                self.terminal_count as usize,
+            ))
+        })?
     }
 
     pub fn write(&self) -> Option<impl Fn(&LimineTerminal, &str)> {
-        let term_func = self.write.get()?.as_ref()?;
+        let term_func = self.write_fn?;
 
-        Some(move |terminal: &LimineTerminal, txt: &str| {
-            term_func.0(terminal as *const _, txt.as_ptr(), txt.len() as u64);
+        Some(move |terminal: &LimineTerminal, txt: &str| unsafe {
+            term_func(terminal as *const _, txt.as_ptr(), txt.len() as u64);
         })
     }
 }
@@ -378,7 +373,7 @@ pub struct LimineSmpResponse {
     pub cpu_count: u64,
     /// Pointer to an array of `cpu_count` pointers to struct [`LimineSmpInfo`]
     /// structures.
-    pub cpus: LiminePtr<*mut LimineSmpInfo>,
+    pub cpus: LiminePtr<LiminePtr<LimineSmpInfo>>,
 }
 
 impl LimineSmpResponse {
@@ -396,8 +391,11 @@ impl LimineSmpResponse {
     /// struct must not be mutated any further.
     pub fn cpus(&mut self) -> Option<&'static mut [LimineSmpInfo]> {
         self.cpus.get().map(|entry| unsafe {
-            core::slice::from_raw_parts_mut(*entry, self.cpu_count as usize)
-        })
+            Some(core::slice::from_raw_parts_mut(
+                entry.as_mut_ptr()?,
+                self.cpu_count as usize,
+            ))
+        })?
     }
 }
 
@@ -447,14 +445,17 @@ pub struct LimineMemmapResponse {
     /// How many memory map entries are present.
     pub entry_count: u64,
     /// Pointer to an array of `entry_count` pointers to struct [`LimineMemmapEntry`] structures.
-    pub entries: LiminePtr<*const LimineMemmapEntry>,
+    pub entries: LiminePtr<LiminePtr<LimineMemmapEntry>>,
 }
 
 impl LimineMemmapResponse {
     pub fn mmap(&self) -> Option<&'static [LimineMemmapEntry]> {
-        self.entries
-            .get()
-            .map(|entry| unsafe { core::slice::from_raw_parts(*entry, self.entry_count as usize) })
+        self.entries.get().map(|entry| unsafe {
+            Some(core::slice::from_raw_parts(
+                entry.as_ptr()?,
+                self.entry_count as usize,
+            ))
+        })?
     }
 }
 
@@ -497,14 +498,17 @@ pub struct LimineModuleResponse {
     /// How many modules are present.
     pub module_count: u64,
     /// Pointer to an array of `module_count` pointers to struct [`LimineFile`] structures
-    pub modules: LiminePtr<*const LimineFile>,
+    pub modules: LiminePtr<LiminePtr<LimineFile>>,
 }
 
 impl LimineModuleResponse {
     pub fn modules(&self) -> Option<&'static [LimineFile]> {
-        self.modules
-            .get()
-            .map(|entry| unsafe { core::slice::from_raw_parts(*entry, self.module_count as usize) })
+        self.modules.get().map(|entry| unsafe {
+            Some(core::slice::from_raw_parts(
+                entry.as_ptr()?,
+                self.module_count as usize,
+            ))
+        })?
     }
 }
 
